@@ -3,12 +3,13 @@
 import click
 import os
 import shutil
-import configparser
+import configparser 
 from datetime import datetime
 import importlib.util
 import importlib.machinery
 from mroll.config import *
 from mroll.migration import Revision, MigrationContext, WorkDirectory
+from mroll.exceptions import RevisionOperationError
 
 
 def load_module_py(module_id, path):
@@ -27,10 +28,27 @@ def get_env():
     env = load_module_py('env', path)
     return env
 
-def rev_id():
+def gen_rev_id():
     import uuid
     return uuid.uuid4().hex[-12:]
-    
+
+def ensure_setup():
+    try:
+        config = Config.from_file(MROLL_CONFIG_FILE)
+        wd = WorkDirectory(config.work_dir)
+        wd.config_validate()
+    except RuntimeError as e:
+        raise SystemExit(e)
+    except ValueError as e:
+        raise SystemExit(e)
+
+def ensure_init():
+    ensure_setup()
+    env = get_env()
+    try:
+        env.get_head()
+    except:
+        raise SystemExit("Error: mroll not initialized! Run init command first.")
 
 # ----------------------------------
 
@@ -95,6 +113,9 @@ def init():
     """
     Creates mroll_revisions tbl. Should be run once.
     """
+    ensure_setup()
+    config = Config.from_file(MROLL_CONFIG_FILE)
+    wd = WorkDirectory(config.work_dir)
     env = get_env()
     try:
         # if following succeeds then mroll revisons tbl exist.
@@ -102,11 +123,8 @@ def init():
         return print("Nothing to do! Mroll revisions table already exist.")
     except:
         pass
-    try:
-        env.create_revisions_table() and print('{} created'.format(env.tbl_name))
-    except Exception as e:
-        print(e)
-        raise SystemExit('init failed')
+    env.create_revisions_table()
+    print('{} table created'.format(env.tbl_name))
     print('Done')
     
 @cli.command(name='revision')
@@ -115,22 +133,16 @@ def revision(message):
     """
     Creates new revision from a template.
     """
+    ensure_setup()
     config = Config.from_file(MROLL_CONFIG_FILE)
     wd = WorkDirectory(config.work_dir)
     ts = datetime.now().isoformat()
-    id_ = rev_id()
+    id_ = gen_rev_id()
     description = message or ''
-    file_ = os.path.join(get_templates_dir(), 'revision_template.txt')
-    with open(file_, 'r') as f:
-        template = f.read()
-        kebab = description.strip().replace(' ', '_')
-        fn = os.path.join(wd.path, 'versions', '{}_{}.sql'.format(id_, kebab))
-        with open(fn, 'w+') as fw:
-            fw.write('-- identifiers used by mroll\n')
-            header = "-- id={}\n-- description={}\n-- ts={}\n".format(id_, description, ts)
-            fw.write(header)
-            fw.write(template)
-        assert os.path.exists(fn)
+    wd.add_revision(Revision(id_, description, ts))
+    kebab = description.strip().replace(' ', '_')
+    fn = os.path.join(wd.path, 'versions', '{}_{}.sql'.format(id_, kebab))
+    assert os.path.exists(fn)
     print('ok')
 
 @cli.command(name='history')
@@ -138,20 +150,27 @@ def history():
     """
     Shows applied revisions.
     """
+    ensure_init()
     return applied_revisions()
     
-def all_revisions():
+def all_revisions(show_patch=False):
     config = Config.from_file(MROLL_CONFIG_FILE)
     wd = WorkDirectory(config.work_dir)
     for rev in wd.revisions:
-        print(rev)
+        if show_patch:
+            print(rev.serialize())
+        else:
+            print(rev)
 
-def applied_revisions():
+def applied_revisions(show_patch=False):
     migr_ctx = MigrationContext.from_env(get_env())
     for r in migr_ctx.revisions:
-        print(r)
+        if show_patch:
+            print(r.serialize())
+        else:
+            print(r)
 
-def pending_revisions():
+def pending_revisions(show_patch=False):
     """
     Shows pending revisions not yet applied.
     """
@@ -165,75 +184,118 @@ def pending_revisions():
             return datetime.fromisoformat(rev.ts) > migr_ctx.head.ts
         working_set = list(filter(filter_fn, working_set))
     for r in working_set:
-        print(r)
+        if show_patch:
+            print(r.serialize())
+        else:
+            print(r)
 
 @cli.command(name="show")
 @click.argument('subcmd', nargs=1)
-def show(subcmd):
+@click.argument('options', required=False)
+def show(subcmd, options):
     """
     Shows revisions information.\n
-    mroll show [ all | pending | applied ]
+    mroll show [ all | pending | applied ] [-p|--patch]
     """
+    ensure_init()
+    def usage(err_msg: str):
+        res ="Error: {}\n".format(err_msg) if err_msg else ''
+        res += 'Usage:\n'
+        res+='mroll show [pending|applied|all] [-p|--patch]\n'
+        return res
+    show_patch = False
     if subcmd not in ['all', 'pending', 'applied']:
-        raise SystemExit("Invalid sub command!")
+        raise SystemExit(usage("Invalid subcommand!"))
+    if options:
+        if options not in ['-p', '--p']:
+            raise SystemExit(usage("Invalid subcommand option!"))
+        show_patch = True 
     if subcmd == 'pending':
-        return pending_revisions()
+        return pending_revisions(show_patch=show_patch)
     if subcmd == 'applied':
-        return applied_revisions()
-    return all_revisions()
+        return applied_revisions(show_patch=show_patch)
+    return all_revisions(show_patch=show_patch)
 
 @cli.command(name="upgrade")
-@click.option('-n', '--num', help="run n number of pending revisions")
-def upgrade(num):
-    """
-    Applies all revisions not yet applied in work dir.
+@click.option('-n', '--num', 'step', help="run n number of pending revisions")
+def upgrade(step):
     """
+    Applies revisions in work dir not yet applied.
+    """ 
+    ensure_init()
     config = Config.from_file(MROLL_CONFIG_FILE)
     wd = WorkDirectory(config.work_dir)
     env = get_env()
     migr_ctx = MigrationContext.from_env(env)
+    # compute working set
     working_set = wd.revisions
     if migr_ctx.head is not None:
-        print('adjusting working set ...')
         def filter_fn(rev):
             return datetime.fromisoformat(rev.ts) > migr_ctx.head.ts
         working_set = list(filter(filter_fn, working_set))
-    ptr = num or len(working_set)
-    for rev in working_set[:ptr]:
-        try:
-            env.add_revision(rev.id, rev.description, rev.ts, rev.upgrade_sql)
-        except Exception as e:
-            print(e)
-            raise SystemExit('Upgrade failed at revision id={} description={}'.format(rev.id, rev.description))
+    ptr = step or len(working_set)
+    # adjust working set
+    working_set = working_set[:ptr]
+    # ensure idempotency
+    for rev in working_set:
+        if rev.upgrade_sql is None:
+            msg="""
+            Error: No upgrade sql script @{}!
+            """.format(rev.id)
+            if rev.downgrade_sql is not None:
+                msg="""
+                Error: No upgrade sql script @{}, while there is a
+                downgrade sql:
+                {}
+                Scripts should be idempotent.
+                """.format(rev.id, rev.downgrade_sql)
+            raise SystemExit(msg)
+    # execute
+    env.add_revisions(working_set)
     print('Done')
 
 @cli.command(name='rollback')
-@click.option('-n', '--num', help="rollbacks n number applied revisions")
-def rollback(num):
+@click.option('-n', '--num', 'step', default=1, help="rollbacks n number applied revisions")
+@click.option('-r', '--rev', 'rev_id', help="rollbacks to specific revision id inclusive")
+def rollback(step, rev_id):
     """
     Downgrades to previous revision by default. 
     """
+    ensure_init()
     config = Config.from_file(MROLL_CONFIG_FILE)
     wd = WorkDirectory(config.work_dir)
-    count = num or 1
-    while count:
-        env = get_env()
-        migr_ctx = MigrationContext.from_env(env)
-        if migr_ctx.head is None:
-            print("Nothing to do!")
-            return
-        print('Rolling back id={} description={} ...'.format(migr_ctx.head.id, migr_ctx.head.description)) 
-        downgrade_sql = ''
-        for rev in reversed(wd.revisions):
-            if rev.id == migr_ctx.head.id:
-                downgrade_sql = rev.downgrade_sql
+    env = get_env()
+    migr_ctx = MigrationContext.from_env(env)
+    if migr_ctx.head is None:
+        raise SystemExit('Nothing to do!')
+    working_set: List[Revision] = list(filter(lambda rev: datetime.fromisoformat(rev.ts) <= migr_ctx.head.ts, wd.revisions))
+    count = 0
+    buff=[]
+    for rev in reversed(working_set):
+        if rev_id is None and count==step:
+            break
+        if rev_id is not None:
+            if rev.id == rev_id:
+                buff.append(rev)
                 break
-        try:
-            env.remove_revision(migr_ctx.head.id, downgrade_sql)
-        except Exception as e:
-            print(e)
-            raise SystemExit('Rollback failed!')
-        count-=1
+        buff.append(rev)
+        count+=1
+    working_set: List[Revision] = [] + buff
+     # insure idempotency
+    for rev in working_set:
+        if rev.downgrade_sql is None:
+            msg="""
+                Error: No downgrade sql script @{}!
+                """.format(rev.id)
+            if rev.upgrade_sql is not None:
+                msg="""
+                Error: No downgrade sql script @{}, while there is a
+                upgrade sql:
+                {}
+                Scripts should be idempotent.
+                """.format(rev.id, rev.upgrade_sql)
+            raise SystemExit(msg)
+    env.remove_revisions(working_set)
     print('Done')
 
 @cli.command(name='version')
